/*
   This file is copyright (c) 1986 by William Couture.  All rights reserved.
*/


#include <compiler.h>
#include <vim.h>
#include <opdefs.h>
#include <regstruc.h>

extern unsigned int clockcount;
extern unsigned int clockdivide;
extern unsigned int hwintflag;
extern unsigned int hwintstack[16];
extern unsigned int rshiftstatus;
extern unsigned int rshiftused;
extern unsigned int lshiftstatus;
extern unsigned int lshiftused;
extern unsigned int keybreak;
extern unsigned int keybdoff;
extern unsigned int hkscancode;
extern unsigned int hotkeypressed;
extern unsigned int port60ptr;
extern unsigned char port60queue[64];
extern unsigned int fake_retrace;
extern unsigned int retr_count;
extern unsigned int retr_num;
extern unsigned int retr_flag;
extern unsigned int cursor_start;
extern unsigned int cursor_flag;
extern unsigned int videobase;
extern unsigned int port64hold;
extern int egascreen;



void int8hdlr()  /* Process the timer interrupt */
   {
#asm
   pushf              ;save current flags, including interrupt flag
   cli                ;give us an interrupt free moment
   push ds            ;save whatever DS was passed us
   push ax            ;save work register AX
;   mov ax,0b800h
;   mov ds,ax
;   inc byte [0]
   mov ax,40h         ;set ax to 40h
   mov ds,ax          ;ds to 40h
   mov byte [40h],255 ;move 255 to the byte at 40:40 (abs 440) so that the
                      ;decrement at the clock tick will NOT turn off the
                      ;floppy motor.  This allows VIM, doing BIOS calls, to
                      ;finish what it is doing before the real BIOS turns
                      ;off the disk drive
   db  2eh            ;SEG CS override
   mov  ds,word [9]   ;get the program's DS, saved at program start
   push si            ;si is used by the following C code
#endasm
   clockcount++;
   if (clockcount > clockdivide)  /* slow down the clock, so that the simulator
                 does not spend all its time servicing the clock interrupt.
                 This gives one clock tick per minute when the simulator is
                 executing.   However, as the simulator is about 1/2000 as
                 fast as the actual 8088, this may have to be changed. */
      {
      hwintflag = TRUE;  /* say an interrupt is waiting */
      hwintstack[8-HWINTBASE] = 8;  /* place Int number in queue */
      clockcount = 0;  /* reset for next count */
      }
#asm
;   mov al,2                ;2
;   sub al,byte cursor_flag_ ;2-flag = 0 or 2
;   mov byte cursor_flag_,al ;save flag for next pass
;   neg al                  ;-2 or 0
;   add al,byte cursor_start_ ;add cursor starting line
;   mov al,byte cursor_flag_ ;get cursor height
;   dec byte cursor_flag_   ;for next time
;   jnz set_cursor_height   ;not zero, no need to reset
;   mov al,byte cursor_start_ ;get cursor max height
;   mov byte cursor_flag_,al  ;set for next time
;set_cursor_height:
;   push	dx                 ;save dx
;   mov dx,word videobase_  ;get base I/O location
;   cmp word egascreen_,0   ;is this an ega screen?
;   jle set_cga_cursor      ;no, skip this
;   mov si,ax               ;save ax
;   mov al,20               ;EGA port address
;   jmp set_cursor_height   ;and skip the next instructin
;set_cga_cursor:
;   mov al,10               ;CGA/MDA port address
;set_cursor_height:
;   out dx,al               ;set register address
;   inc dx                  ;next port
;   mov ax,si               ;get data back
;   jmp waste_time          ;time delay for AT
;waste_time:
;   out dx,al               ;and set cursor height
;   pop dx                  ;restore dx
   cmp word fake_retrace_,1 ;are we doing a fake retrace?
   jne clock_int_1         ;no, skip this
   dec byte retr_count_    ;count this tick
   jnz clock_int_1         ;not zero, not time to change yet
   mov ax,word retr_num_   ;get length of retrace
   mov word retr_count_,ax ;reset counter
   xor byte retr_flag_,1   ;toggle TRUE/FALSE
clock_int_1:
   pop si                  ;restore index register
   pop ax                  ;restore work register
   pop ds                  ;restore DS to it state at routine entry
                           ;flags are still on stack to simulate INT for the
                           ;next procedure in interrupt handler chain
      ;NOTE  lcall word mnemonic is DeSmet specific.
      ;      MASM would be    CALL   DWORD PTR CS:[000bh]  or at least close
      ;      to that.
   db  2eh               ;SEG CS override
   lcall word [000bh]    ;far (long) call to old int routine
                         ;Note that old routine MUST be CALLed, so it can
                         ;return here to
   pop bp                ;restore stack for/from C routine
   iret                  ;return to calling routine
#endasm
   }

void int9hdlr()  /* Use this as an example C code interrupt handler */
   {
#asm
   pushf              ;save current flags, including interrupt flag
   cli                ;give us an interrupt free moment
   push ds            ;save whatever DS was passed us
   db  2eh            ;SEG CS override
   mov  ds,word [9]   ;get the program's DS, saved at program start
   push si            ;si is used by the following C code
   push ax                  ;save ax register
   in al,64h                ;get the AT keyboard status register
   mov byte port64hold_,al  ;save for later
   in  al,60h               ;get scancode
   mov ah,al                ;copy into ah for processing
   or ah,byte keybreak_     ;new keyboard encoding break flag
   cmp ah,36h               ;right shift pressed?
   jne checknotshift        ;no, skip this
   mov byte rshiftstatus_,1 ;set shift status for hotkey
   jmp checkforhotkey       ;go see if this is the hotkey
checknotshift:
   cmp ah,0b6h              ;right shift released?
   jne checkleftshift       ;no, skip clearing flag
   mov byte rshiftstatus_,0 ;clear shift status for hotkey
   mov byte rshiftused_,0   ;signal this shift as not used
   jmp checkforhotkey       ;and see about other stuff
checkleftshift:
   cmp ah,2ah               ;left shift pressed?
   jne checknotleft         ;no, skip this
   mov byte lshiftstatus_,1 ;set left shift status for hotkey
   jmp checkforhotkey       ;and see about hotkey
checknotleft:
   cmp ah,0aah              ;left shift released?
   jne checkforhotkey       ;no, skip clearing flag
   mov byte lshiftstatus_,0 ;clear shift status for hotkey
   mov byte lshiftused_,0   ;signal this shift as not used
checkforhotkey:
   cmp ah,byte hkscancode_  ;is this the "hot key"?
   jne saveport60           ;no, process as normal
   cmp byte rshiftstatus_,1 ;is right shift pressed?
   jne isitleft             ;no, check on left shift
   mov byte hotkeypressed_,1 ;flag interrupt to VIM
   jmp doint9now            ;do not pass hotkey to simulator
isitleft:
   cmp byte lshiftstatus_,1 ;is left shift pressed?
   jne saveport60           ;no, process as normal
   mov byte hotkeypressed_,2 ;flag interrupt to VIM
   jmp doint9now            ;do not pass hotkey to simulator
saveport60:
   cmp byte keybdoff_,1     ;is keyboard off?
   je doint9now             ;yes, screw this
   mov si,word port60ptr_   ;get pointer into keyboard port
   cmp si,60                ;queue full?
   jae doint9now            ;yes, skip processing
   add si,offset port60queue_  ;add array base
   cmp ah,36h               ;right shift pressed? (again)
   je  doint9now            ;yes, do not save it yet
   cmp ah,2ah               ;left shift pressed? (again)
   je  doint9now            ;yes, do not save it yet
   cmp byte rshiftstatus_,1 ;right shift pressed?
   jne saveleft             ;no, skip this
   cmp byte rshiftused_,1   ;has the shift been recorded?
   je  saveleft             ;yes, skip this
   mov byte [si],36h        ;move a right shift into the queue before the
                            ;current character
   inc si                   ;update si for next keypress into buffer
   inc word port60ptr_      ;update pointer into port60queue
   mov byte rshiftused_,1   ;flag this shift as USED
   jmp moresavep60          ;and save the keypress
saveleft:
   cmp byte lshiftstatus_,1 ;left shift pressed?
   jne moresavep60          ;no, skip it
   cmp byte lshiftused_,1   ;has the shift been recorded?
   je  moresavep60          ;yes, skip this
   mov byte [si],2ah        ;move a left shift into the queue before the
                            ;current character
   inc si                   ;update si for next keypress into buffer
   inc word port60ptr_      ;update pointer into port60queue
   mov byte lshiftused_,1   ;and flag this shift as USED
moresavep60:
   mov byte [si],al         ;save this keypress
   inc word port60ptr_      ;and update pointer into port60queue
   mov word hwintflag_,1    ;tell simulator of interrupt
   mov si,2                 ;9-HWINTBASE, *2 for word address
   add si,offset hwintstack_  ;add array base
   mov word [si],9          ;say who is requesting interrupt
doint9now:
   mov byte keybreak_,0     ;default to no breakcode
   cmp al,0f0h              ;is this the new break code?
   jne finish9              ;no, finish interrupt 9
   mov byte keybreak_,80h   ;set flag for next pass
finish9:
   pop ax                   ;restore work register
   pop si                   ;restore index register
   pop ds                   ;restore DS to it state at routine entry
                            ;flags are still on stack to simulate INT for the
                            ;next procedure in interrupt handler chain
      ;NOTE  lcall word mnemonic is DeSmet specific.
      ;      MASM would be    CALL   DWORD PTR CS:[000fh]  or at least close
      ;      to that.
   db  2eh                  ;SEG CS override
   lcall word [000fh]       ;far (long) call to old int routine
                            ;Note that old routine MUST be CALLed, so it can
                            ;return here to
   pop   bp                 ;restore stack for/from C routine
   iret                     ;return to caller
#endasm
   }

void intahdlr()  /* null interrupt handler to point an interrupt to */
   {
#asm
;   pushf                   save current flags, including interrupt flag
   cli                     ;give us an interrupt free moment
   push ds                 ;save whatever DS was passed us
   db  2eh                 ;SEG CS override
   mov  ds,word [9]        ;get the program's DS, saved at program start
   push si                 ;si is used by the following C code
#endasm
   hwintflag = TRUE;  /* say an interrupt is pending */
   hwintstack[0x0a-HWINTBASE] = 0x0a;  /* and place flag in queue */
#asm
;   push ax                 ;save work register AX
   ;
   ;  work routine goes here
   ;
;   pop ax                  ;restore work register
   pop si                  ;restore index register
   pop ds                  ;restore DS to it state at routine entry
                           ;flags are still on stack to simulate INT for the
                           ;next procedure in interrupt handler chain
      ;NOTE  lcall word mnemonic is DeSmet specific.
      ;      MASM would be    CALL   DWORD PTR CS:[0013h]  or at least close
      ;      to that.
;   db  2eh               SEG CS override
;   lcall word [0013h]    far (long) call to old int routine
                         ;Note that old routine MUST be CALLed, so it can
                         ;return here to
   pop bp                ;restore stack for/from C routine
   iret                  ;return to calling routine
#endasm
   }

void intbhdlr()  /* null interrupt handler to point an interrupt to */
   {
#asm
;   pushf                   save current flags, including interrupt flag
   cli                     ;give us an interrupt free moment
   push ds                 ;save whatever DS was passed us
   db  2eh                 ;SEG CS override
   mov  ds,word [9]        ;get the program's DS, saved at program start
   push si                 ;si is used by the following C code
#endasm
   hwintflag = TRUE;  /* say an interrupt is pending */
   hwintstack[0x0b-HWINTBASE] = 0x0b;  /* and place flag in queue */
#asm
;   push ax                 ;save work register AX
   ;
   ;  work routine goes here
   ;
;   pop ax                  ;restore work register
   pop si                  ;restore index register
   pop ds                  ;restore DS to it state at routine entry
                           ;flags are still on stack to simulate INT for the
                           ;next procedure in interrupt handler chain
      ;NOTE  lcall word mnemonic is DeSmet specific.
      ;      MASM would be    CALL   DWORD PTR CS:[0017h]  or at least close
      ;      to that.
;   db  2eh               SEG CS override
;   lcall word [0017h]    far (long) call to old int routine
                         ;Note that old routine MUST be CALLed, so it can
                         ;return here to
   pop bp                ;restore stack for/from C routine
   iret                  ;return to calling routine
#endasm
   }

void intchdlr()  /* null interrupt handler to point an interrupt to */
   {
#asm
;   pushf                   save current flags, including interrupt flag
   cli                     ;give us an interrupt free moment
   push ds                 ;save whatever DS was passed us
   db  2eh                 ;SEG CS override
   mov  ds,word [9]        ;get the program's DS, saved at program start
   push si                 ;si is used by the following C code
#endasm
   hwintflag = TRUE;  /* say an interrupt is pending */
   hwintstack[0x0c-HWINTBASE] = 0x0c;  /* and place flag in queue */
#asm
;   push ax                 ;save work register AX
   ;
   ;  work routine goes here
   ;
;   pop ax                  ;restore work register
   pop si                  ;restore index register
   pop ds                  ;restore DS to it state at routine entry
                           ;flags are still on stack to simulate INT for the
                           ;next procedure in interrupt handler chain
      ;NOTE  lcall word mnemonic is DeSmet specific.
      ;      MASM would be    CALL   DWORD PTR CS:[001bh]  or at least close
      ;      to that.
;   db  2eh               SEG CS override
;   lcall word [001bh]    far (long) call to old int routine
                         ;Note that old routine MUST be CALLed, so it can
                         ;return here to
   pop bp                ;restore stack for/from C routine
   iret                  ;return to calling routine
#endasm
   }

void intdhdlr()  /* null interrupt handler to point an interrupt to */
   {
#asm
;   pushf                   save current flags, including interrupt flag
   cli                     ;give us an interrupt free moment
   push ds                 ;save whatever DS was passed us
   db  2eh                 ;SEG CS override
   mov  ds,word [9]        ;get the program's DS, saved at program start
   push si                 ;si is used by the following C code
#endasm
   hwintflag = TRUE;  /* say an interrupt is pending */
   hwintstack[0x0d-HWINTBASE] = 0x0d;  /* and place flag in queue */
#asm
;   push ax                 ;save work register AX
   ;
   ;  work routine goes here
   ;
;   pop ax                  ;restore work register
   pop si                  ;restore index register
   pop ds                  ;restore DS to it state at routine entry
                           ;flags are still on stack to simulate INT for the
                           ;next procedure in interrupt handler chain
      ;NOTE  lcall word mnemonic is DeSmet specific.
      ;      MASM would be    CALL   DWORD PTR CS:[001fh]  or at least close
      ;      to that.
;   db  2eh               SEG CS override
;   lcall word [001fh]    far (long) call to old int routine
                         ;Note that old routine MUST be CALLed, so it can
                         ;return here to
   pop bp                ;restore stack for/from C routine
   iret                  ;return to calling routine
#endasm
   }

void intehdlr()  /* null interrupt handler to point an interrupt to */
   {
#asm
;   pushf                   ;save current flags, including interrupt flag
   cli                     ;give us an interrupt free moment
   push ds                 ;save whatever DS was passed us
   db  2eh                 ;SEG CS override
   mov  ds,word [9]        ;get the program's DS, saved at program start
   push si                 ;si is used by the following C code
#endasm
   hwintflag = TRUE;  /* say an interrupt is pending */
   hwintstack[0x0e-HWINTBASE] = 0x0e;  /* and place flag in queue */
#asm
;   push ax                 ;save work register AX
   ;
   ;  work routine goes here
   ;
;   pop ax                  ;restore work register
   pop si                  ;restore index register
   pop ds                  ;restore DS to it state at routine entry
                           ;flags are still on stack to simulate INT for the
                           ;next procedure in interrupt handler chain
      ;NOTE  lcall word mnemonic is DeSmet specific.
      ;      MASM would be    CALL   DWORD PTR CS:[0023h]  or at least close
      ;      to that.
;   db  2eh               ;SEG CS override
;   lcall word [0023h]    ;far (long) call to old int routine
                         ;Note that old routine MUST be CALLed, so it can
                         ;return here to
   pop bp                ;restore stack for/from C routine
   iret                  ;return to calling routine
#endasm
   }

void intfhdlr()  /* null interrupt handler to point an interrupt to */
   {
#asm
;   pushf                   save current flags, including interrupt flag
   cli                     ;give us an interrupt free moment
   push ds                 ;save whatever DS was passed us
   db  2eh                 ;SEG CS override
   mov  ds,word [9]        ;get the program's DS, saved at program start
   push si                 ;si is used by the following C code
#endasm
   hwintflag = TRUE;  /* say an interrupt is pending */
   hwintstack[0x0f-HWINTBASE] = 0x0f;  /* and place flag in queue */
#asm
;   push ax                 ;save work register AX
   ;
   ;  work routine goes here
   ;
;   pop ax                  ;restore work register
   pop si                  ;restore index register
   pop ds                  ;restore DS to it state at routine entry
                           ;flags are still on stack to simulate INT for the
                           ;next procedure in interrupt handler chain
      ;NOTE  lcall word mnemonic is DeSmet specific.
      ;      MASM would be    CALL   DWORD PTR CS:[0027h]  or at least close
      ;      to that.
;   db  2eh               SEG CS override
;   lcall word [0027h]    far (long) call to old int routine
                         ;Note that old routine MUST be CALLed, so it can
                         ;return here to
   pop bp                ;restore stack for/from C routine
   iret                  ;return to calling routine
#endasm
   }

void int70hdlr()  /* null interrupt handler to point an interrupt to */
   {
#asm
;   pushf                   save current flags, including interrupt flag
   cli                     ;give us an interrupt free moment
   push ds                 ;save whatever DS was passed us
   db  2eh                 ;SEG CS override
   mov  ds,word [9]        ;get the program's DS, saved at program start
   push si                 ;si is used by the following C code
#endasm
   hwintflag = TRUE;  /* say an interrupt is pending */
   hwintstack[0x70-ATINTOFFSET] = 0x70;  /* and place flag in queue */
#asm
;   push ax                 ;save work register AX
   ;
   ;  work routine goes here
   ;
;   pop ax                  ;restore work register
   pop si                  ;restore index register
   pop ds                  ;restore DS to it state at routine entry
                           ;flags are still on stack to simulate INT for the
                           ;next procedure in interrupt handler chain
      ;NOTE  lcall word mnemonic is DeSmet specific.
      ;      MASM would be    CALL   DWORD PTR CS:[002bh]  or at least close
      ;      to that.
;   db  2eh               SEG CS override
;   lcall word [002bh]    far (long) call to old int routine
                         ;Note that old routine MUST be CALLed, so it can
                         ;return here to
   pop bp                ;restore stack for/from C routine
   iret                  ;return to calling routine
#endasm
   }

void int71hdlr()  /* null interrupt handler to point an interrupt to */
   {
#asm
;   pushf                   save current flags, including interrupt flag
   cli                     ;give us an interrupt free moment
   push ds                 ;save whatever DS was passed us
   db  2eh                 ;SEG CS override
   mov  ds,word [9]        ;get the program's DS, saved at program start
   push si                 ;si is used by the following C code
#endasm
   hwintflag = TRUE;  /* say an interrupt is pending */
   hwintstack[0x71-ATINTOFFSET] = 0x71;  /* and place flag in queue */
#asm
;   push ax                 ;save work register AX
   ;
   ;  work routine goes here
   ;
;   pop ax                  ;restore work register
   pop si                  ;restore index register
   pop ds                  ;restore DS to it state at routine entry
                           ;flags are still on stack to simulate INT for the
                           ;next procedure in interrupt handler chain
      ;NOTE  lcall word mnemonic is DeSmet specific.
      ;      MASM would be    CALL   DWORD PTR CS:[002fh]  or at least close
      ;      to that.
;   db  2eh               SEG CS override
;   lcall word [002fh]    far (long) call to old int routine
                         ;Note that old routine MUST be CALLed, so it can
                         ;return here to
   pop bp                ;restore stack for/from C routine
   iret                  ;return to calling routine
#endasm
   }

void int72hdlr()  /* null interrupt handler to point an interrupt to */
   {
#asm
;   pushf                   save current flags, including interrupt flag
   cli                     ;give us an interrupt free moment
   push ds                 ;save whatever DS was passed us
   db  2eh                 ;SEG CS override
   mov  ds,word [9]        ;get the program's DS, saved at program start
   push si                 ;si is used by the following C code
#endasm
   hwintflag = TRUE;  /* say an interrupt is pending */
   hwintstack[0x72-ATINTOFFSET] = 0x72;  /* and place flag in queue */
#asm
;   push ax                 ;save work register AX
   ;
   ;  work routine goes here
   ;
;   pop ax                  ;restore work register
   pop si                  ;restore index register
   pop ds                  ;restore DS to it state at routine entry
                           ;flags are still on stack to simulate INT for the
                           ;next procedure in interrupt handler chain
      ;NOTE  lcall word mnemonic is DeSmet specific.
      ;      MASM would be    CALL   DWORD PTR CS:[0033h]  or at least close
      ;      to that.
;   db  2eh               SEG CS override
;   lcall word [0033h]    far (long) call to old int routine
                         ;Note that old routine MUST be CALLed, so it can
                         ;return here to
   pop bp                ;restore stack for/from C routine
   iret                  ;return to calling routine
#endasm
   }

void int73hdlr()  /* null interrupt handler to point an interrupt to */
   {
#asm
;   pushf                   save current flags, including interrupt flag
   cli                     ;give us an interrupt free moment
   push ds                 ;save whatever DS was passed us
   db  2eh                 ;SEG CS override
   mov  ds,word [9]        ;get the program's DS, saved at program start
   push si                 ;si is used by the following C code
#endasm
   hwintflag = TRUE;  /* say an interrupt is pending */
   hwintstack[0x73-ATINTOFFSET] = 0x73;  /* and place flag in queue */
#asm
;   push ax                 ;save work register AX
   ;
   ;  work routine goes here
   ;
;   pop ax                  ;restore work register
   pop si                  ;restore index register
   pop ds                  ;restore DS to it state at routine entry
                           ;flags are still on stack to simulate INT for the
                           ;next procedure in interrupt handler chain
      ;NOTE  lcall word mnemonic is DeSmet specific.
      ;      MASM would be    CALL   DWORD PTR CS:[0037h]  or at least close
      ;      to that.
;   db  2eh               SEG CS override
;   lcall word [0037h]    far (long) call to old int routine
                         ;Note that old routine MUST be CALLed, so it can
                         ;return here to
   pop bp                ;restore stack for/from C routine
   iret                  ;return to calling routine
#endasm
   }

void int74hdlr()  /* null interrupt handler to point an interrupt to */
   {
#asm
;   pushf                   save current flags, including interrupt flag
   cli                     ;give us an interrupt free moment
   push ds                 ;save whatever DS was passed us
   db  2eh                 ;SEG CS override
   mov  ds,word [9]        ;get the program's DS, saved at program start
   push si                 ;si is used by the following C code
#endasm
   hwintflag = TRUE;  /* say an interrupt is pending */
   hwintstack[0x74-ATINTOFFSET] = 0x74;  /* and place flag in queue */
#asm
;   push ax                 ;save work register AX
   ;
   ;  work routine goes here
   ;
;   pop ax                  ;restore work register
   pop si                  ;restore index register
   pop ds                  ;restore DS to it state at routine entry
                           ;flags are still on stack to simulate INT for the
                           ;next procedure in interrupt handler chain
      ;NOTE  lcall word mnemonic is DeSmet specific.
      ;      MASM would be    CALL   DWORD PTR CS:[003bh]  or at least close
      ;      to that.
;   db  2eh               SEG CS override
;   lcall word [003bh]    far (long) call to old int routine
                         ;Note that old routine MUST be CALLed, so it can
                         ;return here to
   pop bp                ;restore stack for/from C routine
   iret                  ;return to calling routine
#endasm
   }

void int75hdlr()  /* null interrupt handler to point an interrupt to */
   {
#asm
;   pushf                   save current flags, including interrupt flag
   cli                     ;give us an interrupt free moment
   push ds                 ;save whatever DS was passed us
   db  2eh                 ;SEG CS override
   mov  ds,word [9]        ;get the program's DS, saved at program start
   push si                 ;si is used by the following C code
#endasm
   hwintflag = TRUE;  /* say an interrupt is pending */
   hwintstack[0x75-ATINTOFFSET] = 0x75;  /* and place flag in queue */
#asm
;   push ax                 ;save work register AX
   ;
   ;  work routine goes here
   ;
;   pop ax                  ;restore work register
   pop si                  ;restore index register
   pop ds                  ;restore DS to it state at routine entry
                           ;flags are still on stack to simulate INT for the
                           ;next procedure in interrupt handler chain
      ;NOTE  lcall word mnemonic is DeSmet specific.
      ;      MASM would be    CALL   DWORD PTR CS:[003fh]  or at least close
      ;      to that.
;   db  2eh               SEG CS override
;   lcall word [003fh]    far (long) call to old int routine
                         ;Note that old routine MUST be CALLed, so it can
                         ;return here to
   pop bp                ;restore stack for/from C routine
   iret                  ;return to calling routine
#endasm
   }

void int76hdlr()  /* null interrupt handler to point an interrupt to */
   {
#asm
;   pushf                   save current flags, including interrupt flag
   cli                     ;give us an interrupt free moment
   push ds                 ;save whatever DS was passed us
   db  2eh                 ;SEG CS override
   mov  ds,word [9]        ;get the program's DS, saved at program start
   push si                 ;si is used by the following C code
#endasm
   hwintflag = TRUE;  /* say an interrupt is pending */
   hwintstack[0x76-ATINTOFFSET] = 0x76;  /* and place flag in queue */
#asm
;   push ax                 ;save work register AX
   ;
   ;  work routine goes here
   ;
;   pop ax                  ;restore work register
   pop si                  ;restore index register
   pop ds                  ;restore DS to it state at routine entry
                           ;flags are still on stack to simulate INT for the
                           ;next procedure in interrupt handler chain
      ;NOTE  lcall word mnemonic is DeSmet specific.
      ;      MASM would be    CALL   DWORD PTR CS:[0043h]  or at least close
      ;      to that.
;   db  2eh               SEG CS override
;   lcall word [0043h]    far (long) call to old int routine
                         ;Note that old routine MUST be CALLed, so it can
                         ;return here to
   pop bp                ;restore stack for/from C routine
   iret                  ;return to calling routine
#endasm
   }

void int77hdlr()  /* null interrupt handler to point an interrupt to */
   {
#asm
;   pushf                   save current flags, including interrupt flag
   cli                     ;give us an interrupt free moment
   push ds                 ;save whatever DS was passed us
   db  2eh                 ;SEG CS override
   mov  ds,word [9]        ;get the program's DS, saved at program start
   push si                 ;si is used by the following C code
#endasm
   hwintflag = TRUE;  /* say an interrupt is pending */
   hwintstack[0x77-ATINTOFFSET] = 0x77;  /* and place flag in queue */
#asm
;   push ax                 ;save work register AX
   ;
   ;  work routine goes here
   ;
;   pop ax                  ;restore work register
   pop si                  ;restore index register
   pop ds                  ;restore DS to it state at routine entry
                           ;flags are still on stack to simulate INT for the
                           ;next procedure in interrupt handler chain
      ;NOTE  lcall word mnemonic is DeSmet specific.
      ;      MASM would be    CALL   DWORD PTR CS:[0047h]  or at least close
      ;      to that.
;   db  2eh               SEG CS override
;   lcall word [0047h]    far (long) call to old int routine
                         ;Note that old routine MUST be CALLed, so it can
                         ;return here to
   pop bp                ;restore stack for/from C routine
   iret                  ;return to calling routine
#endasm
   }

void ctrlchdlr()
   {
#asm
   clc       ;continue program execution
   pop bp    ;restore stack for/from C routine
   db 0cbh   ;RET FAR, not supported by ASM88
#endasm
   }
