/*
   This file is copyright (c) 1986 by William Couture.  All rights reserved.
*/


/*
   Note:  All routines in this file that affect memory addresses DO NOTHING
   if memok is FALSE.  MEMORY FETCHES, however, operate normally.  This way,
   if memok is set to FALSE in the middle of an instruction, things halt as
   they were at the start of the instruction, and the instruction and its
   operands are set normally.
*/

#include <compiler.h>
#include <vim.h>
#include <regstruc.h>
#include <opdefs.h>
#include <machine.h>

extern union regunion regs;

extern unsigned int memok;
extern unsigned int eoiwait;
extern unsigned int hw8259on;
extern unsigned int hwintflag;
extern unsigned int hwintstack[16];
extern unsigned int dmamap[16];
extern unsigned int port60ptr;
extern unsigned int portbreaklist[21];
extern unsigned int portbreaknum;
extern unsigned int keybdoff;

extern unsigned char port60queue[64];

extern unsigned int videobase;
extern unsigned int attribflipflop;
extern unsigned char mdaports[8];
extern unsigned char cgaports[6];
extern unsigned char mdaregs[40];
extern unsigned char cgaregs[40];
extern unsigned char siregs[20];
extern unsigned char attribregs[50];
extern unsigned char seqregs[10];
extern unsigned char egaports[28];

extern void (*checkaddress)();
extern void (*checkwaddress)();

void putbyte(seg,ofs,value)  /* put byte into virtual memory */
   unsigned int seg,ofs,value;
   {
   (*checkaddress)(&seg,&ofs);  /* translate to real address */
   if (memok) 
      {
/*      POKE( ((unsigned char)value),seg,ofs);  and put into memory */
#asm
   mov es,word ptr [bp+4]
   mov bx,word ptr [bp+6]
   mov al,byte ptr [bp+8]
   mov byte ptr es:[bx],al
#endasm
      }
   }

void putword(seg1,ofs1,value)  /* as above, but for a word */
   unsigned int seg1,ofs1,value;
   {
   unsigned int seg2,ofs2;

   seg2 = seg1;
   ofs2 = ofs1 + 1;
   (*checkwaddress)(&seg1,&ofs1,&seg2,&ofs2);
   if (memok)
      {
/*      POKE( ((unsigned char) (value & 0xff)),seg,ofs);  low byte
      POKE( ((unsigned char) ((value >> 8) & 0xff)),seg,ofs+1);  high */
#asm
   mov es,word ptr [bp+4]
   mov bx,word ptr [bp+6]
   mov ax,word ptr [bp+8]
;   mov word es:[bx],ax
   mov byte ptr es:[bx],al
   les bx,dword ptr [bp-4]
   mov byte ptr es:[bx],ah
#endasm
      }
   }

unsigned int fetchbyte(seg,ofs)  /* fetch a byte from virtual memory */
   unsigned int seg,ofs;
   {
   (*checkaddress)(&seg,&ofs);  /* Translate to real address.  Note that
        the proper byte is returned, even if this is a protected address. */
/*   return (unsigned int)PEEK(seg,ofs);  and get the byte */
#asm
   mov es,word ptr [bp+4]
   mov bx,word ptr [bp+6]
   xor ax,ax
   mov al,byte ptr es:[bx]
#endasm
   }

unsigned int fetchword(seg1,ofs1)  /* as above, but for words */
   unsigned int seg1,ofs1;
   {
   unsigned int seg2,ofs2;

   seg2 = seg1;
   ofs2 = ofs1 + 1;
   (*checkwaddress)(&seg1,&ofs1,&seg2,&ofs2);
/*   return( PEEK(seg,ofs) + (PEEK(seg,ofs+1) << 8) ); */
#asm
   mov es,word ptr [bp+4]
   mov bx,word ptr [bp+6]
;   mov ax,word ptr es:[bx]
   mov al,byte ptr es:[bx]
   les bx,dword ptr [bp-4]
   mov ah,byte ptr es:[bx]
#endasm
   }

void tostack(value)  /* push word to stack */
   unsigned int value;
   {
   regs.r.sp -= 2;  /* update stack pointer */
   putword(regs.r.ss,regs.r.sp,value);
   if (!memok)
      regs.r.sp += 2;  /* if not ok, restore stack pointer */
   }

unsigned int fromstack()  /* pop word from stack */
   {
   unsigned int temp;

   temp = fetchword(regs.r.ss,regs.r.sp);
   if (memok)  /* if ok, update stack pointer */
      regs.r.sp += 2;
   return(temp);
   }

unsigned int fromqueue()  /* Fetch the next program byte.  This is called
                             from decode() and decodens() instead of
                             fetchbyte() as it is faster, as no parameters
                             are passed, and the segment and offset are
                             known */
   {
   unsigned int seg,ofs;

   seg = regs.r.cs;  /* make copies that we can change */
   ofs = regs.r.ip++;  /* remember to update IP */
   (*checkaddress)(&seg,&ofs);  /* Translate to real address.  Note that
        the proper byte is returned, even if this is a protected address. */
/*   return (unsigned int)PEEK(seg,ofs);  and get the byte */
#asm
   les bx,dword ptr [bp-4]
   xor ax,ax
   mov al,byte ptr es:[bx]
#endasm
   }

void outmap(portnum,value,length)  /* map the values before outing them */
   unsigned int portnum,value,length;
   {
   unsigned int temp1,temp2;

   if (length == BYTE)
      {
      if (adjustout(portnum,&value))
         OUTBPORT(value,portnum);
      }
   else
      {
      temp1 = value;
      temp2 = (value >> 8);
      if ((adjustout(portnum,&temp1)) && (adjustout(portnum+1,&temp2)))
         {
         OUTWPORT((temp2 << 8) + temp1,portnum);
         if ((portnum == 0x3c0) && (temp1 < 0x20))
            {
            attribregs[temp1] = temp2;
            attribflipflop = 1 - attribflipflop;
            }
         }
      }
   }

int adjustout(portnum,value)  /* map the output ports, taking care
                                      of special values */
   unsigned int portnum, *value;
   {
   unsigned int hold,count,ok;

   ok = TRUE;
   hold = *value & 0xff;  /* get the low byte */
   if (portnum == videobase+6)
      attribflipflop = 0;
   if ((portnum >= 0x3b0) && (portnum <= 0x3bf))  /* if MDA video */
      {
      switch(portnum)
         {
         case 0x3b0:  /* the hardware is not exactly specific, is it? */
         case 0x3b2:
         case 0x3b4:
         case 0x3b6:
            mdaports[P3B4] = hold;
            break;
         case 0x3b1:
         case 0x3b3:
         case 0x3b5:
         case 0x3b7:
            if ((count = mdaports[P3B4]) < NUM_CRT_REG)
               mdaregs[count] = hold;
            break;
         case 0x3b8:
         case 0x3b9:
         case 0x3ba:
         case 0x3bb:
            mdaports[P3B8+portnum-0x3b8] = hold;
            if (portnum == 0x3ba)
               egaports[P3xA] = hold;
            break;
         case 0x3bf:
            mdaports[P3BF] = hold;
            break;
         }
      }
 if ((portnum >= 0x3d0) && (portnum <= 0x3db))  /* if CGA video */
      {
      switch(portnum)
         {
         case 0x3d0:
         case 0x3d2:
         case 0x3d4:
         case 0x3d6:
            cgaports[P3D4] = hold;
            break;
         case 0x3d1:
         case 0x3d3:
         case 0x3d5:
         case 0x3d7:
            if ((count = cgaports[P3D4]) < NUM_CRT_REG)
               cgaregs[count] = hold;
            break;
         case 0x3d8:
         case 0x3d9:
         case 0x3da:
         case 0x3db:
            cgaports[P3D8+portnum-0x3d8] = hold;
            if (portnum == 0x3da)
               egaports[P3xA] = hold;
            break;
         }
      }
   if ((portnum >= 0x3c0) && (portnum < 0x3d0))
      {
      switch (portnum)
         {
         case 0x3c0:
            if (attribflipflop)
               {
               if ((count = egaports[P3C0]) < NUM_ATTRIB_REG)
                  attribregs[count] = hold;
               }
            else
               egaports[P3C0] = hold;
            attribflipflop = 1 - attribflipflop;
            break;
         case 0x3c2:
            egaports[P3C2] = hold;
            break;
         case 0x3c3:
            egaports[P3C3] = hold;
            break;
         case 0x3c4:
            egaports[P3C4] = hold;
            break;
         case 0x3c5:
            if ((count = egaports[P3C4]) < NUM_SEQ_REG)
               seqregs[count] = hold;
            break;
         case 0x3ca:
            egaports[P3CA] = hold;
            break;
         case 0x3cc:
            egaports[P3CC] = hold;
            break;
         case 0x3ce:
            egaports[P3CE] = hold;
            break;
         case 0x3cf:
            if ((count = egaports[P3CE]) < NUM_SI_REG)
               siregs[count] = hold;
            break;
         }
      }
   if (portnum == 0x21)
      hold &= 0xfd;  /* mask off IRQ1 (keyboard interrupt) flag, so that
                        you cannot return to VIM with keyboard dead */
   if ((portnum == 0x61) && (hold & 0x80))
      {
      eoiwait = 2;  /* wait for ints to be on again before
                       queueing the next keycode */
      port60queue[0] = '\0';
      hold &= 0x7f;  /* This turns off bit 7, so that
                        the keyboard is never really turned off
                        by the simulator. */
      }
   if ((portnum == 0x20) && ((hold == 0x20) || ((hold & 0xf0) == 0x60)))
      {
      hw8259on = TRUE;  /* "re-enable" interrupts */
      if (eoiwait == 2)  /* if waiting for keyboard again */
         {
         eoiwait = 0;  /* turn keyboard back on */
         if (port60ptr)  /* if queue not empty */
            {
            port60ptr--;
            if (port60ptr)  /* if still not empty */
               {
               for (count = 0; count < port60ptr; count++)
                  port60queue[count] = port60queue[count+1];
               hwintflag = TRUE;
               hwintstack[9-HWINTBASE] = 9;
               }
            }
         }
      }

#ifdef NEVERDEF
   if ((portnum == 0x64) && (hold == 0xae)
                   && (eoiwait == 1))
      {
      eoiwait = 0;
      if (port60ptr)  /* if queue not empty */
         {
         port60ptr--;
         if (port60ptr)  /* if still not empty */
            {
            for (count = 0; count < port60ptr; count++)
               port60queue[count] = port60queue[count+1];
            hwintflag = TRUE;
            hwintstack[9-HWINTBASE] = 9;
            }
         }
      }
#endif

   if ( ((portnum >= 0x81) && (portnum <= 0x83))
                 || ((portnum >= 0x89) && (portnum <= 0x8b)) )
      hold = (hold & 0x0f)+dmamap[hold & 0x0f];
   if ((portnum == 0x64) && (hold == 0xad))
            /* AT keyboard OFF */
      {
      eoiwait = 2;   /* at next EOI, update keyboard queue */
      ok = FALSE;
      }
   *value = hold;
   return(ok);
   }

void checkport(size,portnum)
   unsigned int size,portnum;
   {
   int i;

   if (portbreaknum)
      {
      for (i = 1; i <= portbreaknum; i++)
         {
         if (portbreaklist[i] == portnum)
            memok = FALSE;
         if (size == WORD)
            if (portbreaklist[i] == portnum+1)
               memok = FALSE;
         }
      }
   else
      memok = FALSE;
   }
