/*
   This file is copyright (c) 1986 by William Couture.  All rights reserved.
*/


#include <compiler.h>
#include <vim.h>
#include <opdefs.h>

/*
   Data for decoding/encoding the 8088 instruction set.  To see where the
   tables come from, see the 8086 & 8088 instant reference card from
   Micro Logic Corp. (more details in comments.h).
*/

/*
   Note:  The usual check for BYTE/WORD instruction is:
      if (instruction.flags & BYTE)
   The lack of a BYTE flag is taken to mean a WORD flag.
*/

struct optable firstbyte[256] = {
{ADD,BYTE,DM,S8},{ADD,WORD,DM,S16},{ADD,BYTE,D8,SM},{ADD,WORD,D16,SM},
{ADD,ACCUM8+DATA8,DRAL,SDATA8},{ADD,ACCUM16+DATA16,DRAX,SDATA16},{PUSH,NONE,DZ,SES},
                                                   {POP,NONE,DES,SZ},
{OR,BYTE,DM,S8},{OR,WORD,DM,S16},{OR,BYTE,D8,SM},{OR,WORD,D16,SM},
{OR,ACCUM8+DATA8,DRAL,SDATA8},{OR,ACCUM16+DATA16,DRAX,SDATA16},{PUSH,NONE,DZ,SCS},
                                                   {UNDEF,NONE,DZ,SZ},
                           /* for 8088 mode only   {POP,NONE,DCS,SZ}, */

{ADC,BYTE,DM,S8},{ADC,WORD,DM,S16},{ADC,BYTE,D8,SM},{ADC,WORD,D16,SM},
{ADC,ACCUM8+DATA8,DRAL,SDATA8},{ADC,ACCUM16+DATA16,DRAX,SDATA16},{PUSH,NONE,DZ,SSS},
                                                   {POP,NONE,DSS,SZ},
{SBB,BYTE,DM,S8},{SBB,WORD,DM,S16},{SBB,BYTE,D8,SM},{SBB,WORD,D16,SM},
{SBB,ACCUM8+DATA8,DRAL,SDATA8},{SBB,ACCUM16+DATA16,DRAX,SDATA16},{PUSH,NONE,DZ,SDS},
                                                   {POP,NONE,DDS,SZ},

{AND,BYTE,DM,S8},{AND,WORD,DM,S16},{AND,BYTE,D8,SM},{AND,WORD,D16,SM},
{AND,ACCUM8+DATA8,DRAL,SDATA8},{AND,ACCUM16+DATA16,DRAX,SDATA16},{SEGES,PREFIX,DZ,SZ},
                                                   {DAA,NONE,DZ,SZ},
{SUB,BYTE,DM,S8},{SUB,WORD,DM,S16},{SUB,BYTE,D8,SM},{SUB,WORD,D16,SM},
{SUB,ACCUM8+DATA8,DRAL,SDATA8},{SUB,ACCUM16+DATA16,DRAX,SDATA16},{SEGCS,PREFIX,DZ,SZ},
                                                   {DAS,NONE,DZ,SZ},

{XOR,BYTE,DM,S8},{XOR,WORD,DM,S16},{XOR,BYTE,D8,SM},{XOR,WORD,D16,SM},
{XOR,ACCUM8+DATA8,DRAL,SDATA8},{XOR,ACCUM16+DATA16,DRAX,SDATA16},{SEGSS,PREFIX,DZ,SZ},
                                                   {AAA,NONE,DZ,SZ},
{CMP,BYTE,DM,S8},{CMP,WORD,DM,S16},{CMP,BYTE,D8,SM},{CMP,WORD,D16,SM},
{CMP,ACCUM8+DATA8,DRAL,SDATA8},{CMP,ACCUM16+DATA16,DRAX,SDATA16},{SEGDS,PREFIX,DZ,SZ},
                                                   {AAS,NONE,DZ,SZ},

/* 1/4 done */

{INC,NONE,DAX,SZ},{INC,NONE,DCX,SZ},{INC,NONE,DDX,SZ},{INC,NONE,DBX,SZ},
{INC,NONE,DSP,SZ},{INC,NONE,DBP,SZ},{INC,NONE,DSI,SZ},{INC,NONE,DDI,SZ},
{DEC,NONE,DAX,SZ},{DEC,NONE,DCX,SZ},{DEC,NONE,DDX,SZ},{DEC,NONE,DBX,SZ},
{DEC,NONE,DSP,SZ},{DEC,NONE,DBP,SZ},{DEC,NONE,DSI,SZ},{DEC,NONE,DDI,SZ},

{PUSH,NONE,DZ,SAX},{PUSH,NONE,DZ,SCX},{PUSH,NONE,DZ,SDX},{PUSH,NONE,DZ,SBX},
{PUSH,NONE,DZ,SSP},{PUSH,NONE,DZ,SBP},{PUSH,NONE,DZ,SSI},{PUSH,NONE,DZ,SDI},
{POP,NONE,DAX,SZ},{POP,NONE,DCX,SZ},{POP,NONE,DDX,SZ},{POP,NONE,DBX,SZ},
{POP,NONE,DSP,SZ},{POP,NONE,DBP,SZ},{POP,NONE,DSI,SZ},{POP,NONE,DDI,SZ},

{PUSHA,NONE,DZ,SZ},{POPA,NONE,DZ,SZ},{BOUND,NONE,D16,SM},{UNDEF,NONE,DZ,SZ},
{UNDEF,NONE,DZ,SZ},{UNDEF,NONE,DZ,SZ},{UNDEF,NONE,DZ,SZ},{UNDEF,NONE,DZ,SZ},
{PUSH,NONE,DZ,SDATA16},{IMUL,DATA16,D16,SM},{PUSH,NONE,DZ,SXDATA8},
                                                   {IMUL,DATA8,D16,SM},
{INSB,BYTE,DZ,SZ},{INSW,WORD,DZ,SZ},{OUTSB,BYTE,DZ,SZ},{OUTSW,WORD,DZ,SZ},

{JO,OFF8,DZ,SOFF8},{JNO,OFF8,DZ,SOFF8},{JNAE,OFF8,DZ,SOFF8},{JAE,OFF8,DZ,SOFF8},
{JZ,OFF8,DZ,SOFF8},{JNZ,OFF8,DZ,SOFF8},{JNA,OFF8,DZ,SOFF8},{JA,OFF8,DZ,SOFF8},
{JS,OFF8,DZ,SOFF8},{JNS,OFF8,DZ,SOFF8},{JPE,OFF8,DZ,SOFF8},{JPO,OFF8,DZ,SOFF8},
{JNGE,OFF8,DZ,SOFF8},{JGE,OFF8,DZ,SOFF8},{JNG,OFF8,DZ,SOFF8},{JG,OFF8,DZ,SOFF8},

/* 1/2 done */

{UNKNOWN,UNK,DA,SM},{UNKNOWN,UNK,DB,SM},{UNKNOWN,UNK,DA,SM},{UNKNOWN,UNK,DC,SM},
{TEST,BYTE,D8,SM},{TEST,WORD,D16,SM},{XCHG,BYTE,D8,SM},{XCHG,WORD,D16,SM},
{MOV,BYTE,DM,S8},{MOV,WORD,DM,S16},{MOV,BYTE,D8,SM},{MOV,WORD,D16,SM},
{MOV,WORD,DM,SL},{LEA,WORD,D16,SM},{MOV,WORD,DL,SM},{POP,WORD,DM,SZ},

{NOP,NONE,DZ,SZ},{XCHG,NONE,DAX,SCX},{XCHG,NONE,DAX,SDX},{XCHG,NONE,DAX,SBX},
{XCHG,NONE,DAX,SSP},{XCHG,NONE,DAX,SBP},{XCHG,NONE,DAX,SSI},{XCHG,NONE,DAX,SDI},
{CBW,NONE,DZ,SZ},{CWD,NONE,DZ,SZ},{CALLF,ADDR32,DZ,SADDR32},{WAIT,NONE,DZ,SZ},
{PUSHF,NONE,DZ,SZ},{POPF,NONE,DZ,SZ},{SAHF,NONE,DZ,SZ},{LAHF,NONE,DZ,SZ},

{MOV,ACCUM8,DRAL,SADDR16},{MOV,ACCUM16,DRAX,SADDR16},{MOV,ACCUM8,DADDR16,SRAL},
                                                   {MOV,ACCUM16,DADDR16,SRAX},
{MOVSB,BYTE,DZ,SZ},{MOVSW,WORD,DZ,SZ},{CMPSB,BYTE,DZ,SZ},{CMPSW,WORD,DZ,SZ},
{TEST,ACCUM8+DATA8,DRAL,SDATA8},{TEST,ACCUM16+DATA16,DRAX,SDATA16},{STOSB,BYTE,DZ,SZ},
                                                   {STOSW,WORD,DZ,SZ},
{LODSB,BYTE,DZ,SZ},{LODSW,WORD,DZ,SZ},{SCASB,BYTE,DZ,SZ},{SCASW,WORD,DZ,SZ},

{MOV,BYTE+DATA8,DAL,SDATA8},{MOV,BYTE+DATA8,DCL,SDATA8},{MOV,BYTE+DATA8,DDL,SDATA8},
                                                   {MOV,BYTE+DATA8,DBL,SDATA8},
{MOV,BYTE+DATA8,DAH,SDATA8},{MOV,BYTE+DATA8,DCH,SDATA8},{MOV,BYTE+DATA8,DDH,SDATA8},
                                                   {MOV,BYTE+DATA8,DBH,SDATA8},
{MOV,WORD+DATA16,DAX,SDATA16},{MOV,WORD+DATA16,DCX,SDATA16},{MOV,WORD+DATA16,DDX,SDATA16},
                                                   {MOV,WORD+DATA16,DBX,SDATA16},
{MOV,WORD+DATA16,DSP,SDATA16},{MOV,WORD+DATA16,DBP,SDATA16},{MOV,WORD+DATA16,DSI,SDATA16},
                                                   {MOV,WORD+DATA16,DDI,SDATA16},

/* 3/4 done */

{UNKNOWN,UNK,DS,SM},{UNKNOWN,UNK,DT,SM},{RET,ADJSTK,DZ,SDATA16},{RET,NONE,DZ,SZ},
{LES,WORD,D16,SM},{LDS,WORD,D16,SM},{MOV,BYTE+DATA8,DM,SDATA8},{MOV,WORD+DATA16,DM,SDATA16},
{ENTER,DATA16+DATA8,DDATA16,SDATA8},{LEAVE,NONE,DZ,SZ},{RETF,ADJSTK,DZ,SDATA16},
                                                   {RETF,NONE,DZ,SZ},
{INT,INT3,DZ,SZ},{INT,NONE,DZ,SDATA8},{INTO,NONE,DZ,SZ},{IRET,NONE,DZ,SZ},

{UNKNOWN,UNK,DD,SM},{UNKNOWN,UNK,DE,SM},{UNKNOWN,UNK,DF,SM},{UNKNOWN,UNK,DG,SM},
{AAM,AADAAM,DZ,SZ},{AAD,AADAAM,DZ,SZ},{UNDEF,NONE,DZ,SZ},{XLAT,BYTE,DZ,SZ},
{ESC,NONE,DM,SZ},{ESC,NONE,DM,SZ},{ESC,NONE,DM,SZ},{ESC,NONE,DM,SZ},
{ESC,NONE,DM,SZ},{ESC,NONE,DM,SZ},{ESC,NONE,DM,SZ},{ESC,NONE,DM,SZ},

{LOOPNZ,OFF8,DZ,SOFF8},{LOOPZ,OFF8,DZ,SOFF8},{LOOP,OFF8,DZ,SOFF8},
                                                   {JCXZ,OFF8,DZ,SOFF8},
{IN,ACCUM8+DATA8,DRAL,SDATA8},{IN,ACCUM16+DATA8,DRAX,SDATA8},{OUT,ACCUM8+DATA8,DDATA8,SRAL},
                                                   {OUT,ACCUM16+DATA8,DDATA8,SRAX},
{CALL,OFF16,DZ,SOFF16},{JMP,OFF16,DZ,SOFF16},{JMPF,ADDR32,DZ,SADDR32},
                                                   {JMP,OFF8,DZ,SOFF8},
{IN,ACCUM8,DRAL,SDX},{IN,ACCUM16,DRAX,SDX},{OUT,ACCUM8,DDX,SRAL},
                                                   {OUT,ACCUM16,DDX,SRAX},

{LOCK,PREFIX,DZ,SZ},{UNDEF,NONE,DZ,SZ},{REPNZ,PREFIX,DZ,SZ},{REPZ,PREFIX,DZ,SZ},
{HLT,NONE,DZ,SZ},{CMC,NONE,DZ,SZ},{UNKNOWN,UNK,DH,SM},{UNKNOWN,UNK,DI,SM},
{CLC,NONE,DZ,SZ},{STC,NONE,DZ,SZ},{CLI,NONE,DZ,SZ},{STI,NONE,DZ,SZ},
{CLD,NONE,DZ,SZ},{STD,NONE,DZ,SZ},{UNKNOWN,UNK,DJ,SM},{UNKNOWN,UNK,DK,SM}

/* and done! */

   };

struct optable secondbyte[13][8] = {
   {
{ADD,BYTE+DATA8,DM,SDATA8},{OR,BYTE+DATA8,DM,SDATA8},{ADC,BYTE+DATA8,DM,SDATA8},
                                                  {SBB,BYTE+DATA8,DM,SDATA8},
{AND,BYTE+DATA8,DM,SDATA8},{SUB,BYTE+DATA8,DM,SDATA8},{XOR,BYTE+DATA8,DM,SDATA8},
                                                  {CMP,BYTE+DATA8,DM,SDATA8}
   },{
{ADD,WORD+DATA16,DM,SDATA16},{OR,WORD+DATA16,DM,SDATA16},{ADC,WORD+DATA16,DM,SDATA16},
                                                  {SBB,WORD+DATA16,DM,SDATA16},
{AND,WORD+DATA16,DM,SDATA16},{SUB,WORD+DATA16,DM,SDATA16},{XOR,WORD+DATA16,DM,SDATA16},
                                                  {CMP,WORD+DATA16,DM,SDATA16}
   },{
{ADD,WORD+XDATA8,DM,SXDATA8},{OR,WORD+XDATA8,DM,SXDATA8},{ADC,WORD+XDATA8,DM,SXDATA8},
                                                  {SBB,WORD+XDATA8,DM,SXDATA8},
{AND,WORD+XDATA8,DM,SXDATA8},{SUB,WORD+XDATA8,DM,SXDATA8},{XOR,WORD+XDATA8,DM,SXDATA8},
                                                  {CMP,WORD+XDATA8,DM,SXDATA8}
   },{
{ROL,NOCOUNT+BYTE,DM,SZ},{ROR,NOCOUNT+BYTE,DM,SZ},{RCL,NOCOUNT+BYTE,DM,SZ},
                                                  {RCR,NOCOUNT+BYTE,DM,SZ},
{SHL,NOCOUNT+BYTE,DM,SZ},{SHR,NOCOUNT+BYTE,DM,SZ},{UNDEF,NONE,DZ,SZ},
                                                  {SAR,NOCOUNT+BYTE,DM,SZ}
   },{
{ROL,NOCOUNT+WORD,DM,SZ},{ROR,NOCOUNT+WORD,DM,SZ},{RCL,NOCOUNT+WORD,DM,SZ},
                                                  {RCR,NOCOUNT+WORD,DM,SZ},
{SHL,NOCOUNT+WORD,DM,SZ},{SHR,NOCOUNT+WORD,DM,SZ},{UNDEF,NONE,DZ,SZ},
                                                  {SAR,NOCOUNT+WORD,DM,SZ}
   },{
{ROL,BYTE+COUNT,DM,SZ},{ROR,BYTE+COUNT,DM,SZ},{RCL,BYTE+COUNT,DM,SZ},{RCR,BYTE+COUNT,DM,SZ},
{SHL,BYTE+COUNT,DM,SZ},{SHR,BYTE+COUNT,DM,SZ},{UNDEF,NONE,DZ,SZ},{SAR,BYTE+COUNT,DM,SZ}
   },{
{ROL,WORD+COUNT,DM,SZ},{ROR,WORD+COUNT,DM,SZ},{RCL,WORD+COUNT,DM,SZ},{RCR,WORD+COUNT,DM,SZ},
{SHL,WORD+COUNT,DM,SZ},{SHR,WORD+COUNT,DM,SZ},{UNDEF,NONE,DZ,SZ},{SAR,WORD+COUNT,DM,SZ}
   },{
{TEST,BYTE+DATA8,DM,SDATA8},{UNDEF,NONE,DZ,SZ},{NOT,BYTE,DM,SZ},{NEG,BYTE,DM,SZ},
{MUL,BYTE,DZ,SM},{IMUL,BYTE,DZ,SM},{DIV,BYTE,DZ,SM},{IDIV,BYTE,DZ,SM}
   },{
{TEST,WORD+DATA16,DM,SDATA16},{UNDEF,NONE,DZ,SZ},{NOT,WORD,DM,SZ},{NEG,WORD,DM,SZ},
{MUL,WORD,DZ,SM},{IMUL,WORD,DZ,SM},{DIV,WORD,DZ,SM},{IDIV,WORD,DZ,SM}
   },{
{INC,BYTE,DM,SZ},{DEC,BYTE,DM,SZ},{UNDEF,NONE,DZ,SZ},{UNDEF,NONE,DZ,SZ},
{UNDEF,NONE,DZ,SZ},{UNDEF,NONE,DZ,SZ},{UNDEF,NONE,DZ,SZ},{UNDEF,NONE,DZ,SZ}
   },{
{INC,WORD,DM,SZ},{DEC,WORD,DM,SZ},{CALL,NONE,DZ,SM},{CALLF,ADDR32,DZ,SM},
{JMP,NONE,DZ,SM},{JMPF,ADDR32,DZ,SM},{PUSH,NONE,DZ,SM},{UNDEF,NONE,DZ,SZ}
   },{
{ROL,BYTE,DM,SDATA8},{ROR,BYTE,DM,SDATA8},{RCL,BYTE,DM,SDATA8},
                                                  {RCR,BYTE,DM,SDATA8},
{SHL,BYTE,DM,SDATA8},{SHR,BYTE,DM,SDATA8},{UNDEF,NONE,DZ,SZ},
                                                  {SAR,BYTE,DM,SDATA8}
   },{
{ROL,WORD,DM,SDATA8},{ROR,WORD,DM,SDATA8},{RCL,WORD,DM,SDATA8},
                                                  {RCR,WORD,DM,SDATA8},
{SHL,WORD,DM,SDATA8},{SHR,WORD,DM,SDATA8},{UNDEF,NONE,DZ,SZ},
                                                  {SAR,WORD,DM,SDATA8}
   }
};
