/*
   This file is copyright (c) 1986 by William Couture.  All rights reserved.
*/


/*
   General header file associated with the files in the VIM program.

   Note that STDIO is not used.  Place all such common definitions
   in this file instead.
*/

/* masks to set appropriate flags in the flags register by ORing */
#define CARRY 0x0001
#define PARITY 0x0004
#define AUXCARRY 0x0010
#define ZERO 0x0040
#define SIGN 0x0080
#define TRAP 0x0100
#define INTERRUPT 0x0200
#define DIRECTION 0x0400
#define OVERFLOW 0x0800

/* masks to clear appropriate flags in the flags register by ANDing */
#define CARRYOFF 0xfffe
#define PARITYOFF 0xfffb
#define AUXCARRYOFF 0xffef
#define ZEROOFF 0xffbf
#define SIGNOFF 0xff7f
#define TRAPOFF 0xfeff
#define INTERRUPTOFF 0xfdff
#define DIRECTIONOFF 0xfbff
#define OVERFLOWOFF 0xf7ff

#define SIGNEDB 0x80  /* mask for sign bit in 8 bit quantity */
#define SIGNEDW 0x8000  /* mask for sign bit in 16 bit quantity */

#define MEMORY 0x0000  /* source/destination is memory */
#define REGISTER 0x0001  /* source/destination is a register */
#define SEGREG 0x0002  /* source/destination is a segment register */
#define REG 0x0003  /* source/destination is a register or segment register */
   /* Note:  I seem to have used REG a lot where I probably should have used
      REGISTER (in fns.c).  Everything seems to work, though. */

#define ESSEG 0x0001  /* prefix is ES segment override */
#define CSSEG 0x0002  /* prefix is CS segment override */
#define SSSEG 0x0004  /* prefix is SS segment override */
#define DSSEG 0x0008  /* prefix is DS segment override */
#define SEG   0x000f  /* prefix is one of the segment overrides */
#define REPZERO  0x0010  /* prefix is REPZ/REPE */
#define REPNOTZERO 0x0020  /* prefix is REPNZ/REPNE */
#define REP   0x0030  /* prefix is one of the REP instructions */
#define FORGETZF 0x0100  /* Set only by string primitive instruction in
                            fns.c, says what this primitive does with a
                            REPxx instruction */
#define ISZFON 0x0200  /* as above for FORGETZF */
#define ISZFOFF 0x0400  /* as above for FORGETZF */

#define HALTERR 1  /* just executed a HALT instruction */
#define NOHALT 0

#define P3B4 0  /* define video port addresses in the mdaports array */
#define P3B5 1
#define P3B8 2
#define P3B9 3
#define P3BA 4
#define P3BB 5
#define P3BF 6

#define P3D4 0  /* define video port addresses in the cgaports array */
#define P3D5 1
#define P3D8 2
#define P3D9 3
#define P3DA 4
#define P3DB 5

#define P3C0 0  /* define the video port addresses in the egaports array */
#define P3C1 1
#define P3C2 2
#define P3C3 3
#define P3C4 4
#define P3C5 5
#define P3C6 6
#define P3C7 7
#define P3C8 8
#define P3C9 9
#define P3CA 10
#define P3CB 11
#define P3CC 12
#define P3CD 13
#define P3CE 14
#define P3CF 15
#define P3xA 16
#define NORMAL_P3C0 17
#define NORMAL_P3C2 18
#define NORMAL_P3C3 19
#define NORMAL_P3C4 20
#define NORMAL_P3C6 21
#define NORMAL_P3CB 22
#define NORMAL_P3CD 23
#define NORMAL_P3CE 24
#define NORMAL_P3BF 25
#define NORMAL_P3x4 26

#define NORM_SEQ_REG 4
#define NORM_SI_REG 8
#define NORM_ATTRIB_REG 19
#define NORM_CRT_REG 24

#define NUM_SEQ_REG 10
#define NUM_SI_REG 20
#define NUM_ATTRIB_REG 50
#define NUM_CRT_REG 40

#define TRUE 1  /* standard true */
#define FALSE 0  /* standard false */

#define YES 1  /* convenient definitions */
#define NO 0

#define HWINTBASE 8  /* lowest numbered hardware interrupt */
#define ATINTBASE 0x70  /* the lowest numbered new AT interrupt */
#define ATINTOFFSET 104 /* 0x68, AT interrupts - HWINTBASE (used to
                           position the flag in hwintstack) */

#define ESC_KEY '\033'  /* the ESC key */

#define HELPSIZE 248  /* number of lines in the "help" command */

/*
   Structure definitions
*/

struct optable  /*  Define the optable structure */
   {
   int name;  /* "name" is also location in arrays */
   unsigned int flags;  /* conditions associated with this instruction */
   int dest;  /* to where (mem/reg/etc) */
   int source;  /* from where (mem/reg/etc) */
   };


struct tofrom  /* stuff associated with instruction destination and source */
   {
   UNSIGNED_LONG value;  /* value of source/destination */
   UNSIGNED_LONG oldvalue;  /* value before it is changed by instruction,
                               used to set flags */
       /* Note:  The file COMPILER.H must be included BEFORE VIM.H, so that
          UNSIGNED_LONG has been TYPEDEFed */
   unsigned int where;  /* is the source/destination register or memory? */
   unsigned int reg;  /* register for source/destination */
   unsigned int seg;  /* segment for souce/destination */
   unsigned int ofs;  /* offset for source/destination */
   };

struct addrtrap            /* structure of the memory address trap info */
   {
   UNSIGNED_LONG lowaddr;  /* the lower bound */
   UNSIGNED_LONG highaddr;  /* the upper bound */
   };

struct intrptr        /* interrupt pointer, for internal use to watch
                         the hardware interrupts */
   {
   unsigned int ptrofs;
   unsigned int ptrseg;
   };
